try:
    from langchain.llms.base import LLM
except ImportError:
    raise ImportError(
        'To use the llm_rs.langchain module, please install llm-rs with the additional "langchain" dependencies via: pip install llm-rs[langchain]')

from typing import Any, Dict, Optional, Sequence, Union, List
import os

from pydantic import root_validator
from langchain.callbacks.manager import CallbackManagerForLLMRun

from ..auto import AutoModel, KnownModels
from ..config import GenerationConfig, SessionConfig
from ..base_model import Model

class RustformerLLM(LLM):
    """
    Langchain-Wrapper around a Rustformers model.
    """

    model: Optional[Model] = None #: :meta private:

    model_path_or_repo_id: Union[str,os.PathLike]
    """The path to the model file or directory or the name of a Hugging Face Hub
    model repo."""

    model_type: Optional[KnownModels] = None
    """The model type."""

    model_file: Optional[str] = None
    """The name of the model file in repo or directory."""

    # session_config:SessionConfig=SessionConfig()
    # """Session config for the model."""

    # generation_config:GenerationConfig=GenerationConfig()
    # """Generation config for the model."""

    lora_paths:Optional[List[Union[str,os.PathLike]]]=None
    """Paths to the lora files."""


    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Get the identifying parameters."""
        return {
            'model_path_or_repo_id': self.model_path_or_repo_id,
            'model_type': self.model_type,
            'model_file': self.model_file,
            # 'session_config': self.session_config,
            # 'generation_config': self.generation_config,
            'lora_paths': self.lora_paths,
        }

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return 'rustformer'

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate and load model from a local file or remote repo."""
        values['model'] = AutoModel.from_pretrained(
            model_path_or_repo_id= values['model_path_or_repo_id'],
            model_type=values['model_type'],
            model_file=values['model_file'],
            # session_config=values['session_config'],
            lora_paths=values['lora_paths'],
        )
        return values

    def _call(
        self,
        prompt: str,
        stop: Optional[Sequence[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
    ) -> str:
        """Generate text from a prompt.

        Args:
            prompt: The prompt to generate text from.
            stop: A list of sequences to stop generation when encountered.

        Returns:
            The generated text.
        """
        text = []
        generation_config = GenerationConfig()

        if stop:
            generation_config.stop_words = list(stop)
        for chunk in self.model.stream(prompt, generation_config=generation_config):
            text.append(chunk)
            if run_manager:
                run_manager.on_llm_new_token(chunk, verbose=self.verbose)
        return ''.join(text)


from typing import Optional,List
from enum import Enum, auto

class ContainerType(Enum):
    GGML=auto(),
    GGJT=auto(),

class QuantizationType(Enum):
    Q4_0=auto(),
    Q4_1=auto(),
    Q5_0=auto(),
    Q5_1=auto(),
    Q8_0=auto(),
    F16=auto(),

class Precision(Enum):
    FP32=auto(),
    FP16=auto(),
    
class GenerationConfig():
    """
    Configuration parameters for the generation process.
    """
    top_k:int
    top_p:float
    temperature:float
    repetition_penalty:float
    repetition_penalty_last_n:int
    seed:int
    max_new_tokens:Optional[int]
    stop_words:Optional[List[str]]
    
    def  __init__(self,top_k:int=40,top_p:float=0.95,temperature:float=0.8,repetition_penalty:float=1.3,repetition_penalty_last_n:int=512,seed:int=42,max_new_tokens:Optional[int]=None,stop_words:Optional[List[str]]=None) -> None: ...
    
class SessionConfig():
    """
    Configure the generation session that will be used for all generations.
    """
    threads:int
    batch_size:int
    keys_memory_type:Precision
    values_memory_type:Precision
    
    
    @property
    def context_length(self)->int: ...
    
    @property
    def prefer_mmap(self)->bool:...
    
    def  __init__(self,threads:int=8,batch_size:int=8,context_length:int=2048,keys_memory_type:Precision=Precision.FP32,values_memory_type:Precision=values_memory_type.FP32,prefer_mmap:bool=True) -> None: ...


from enum import Enum

class StopReason(Enum):
    """
    The reason why generation was stopped.
    """
    EndToken=0,
    MaxLength=1,
    UserCancelled=2

class GenerationTimes():
    """
    Contains the time taken for each step of generation in milliseconds.
    """
    @property
    def total(self)->int: ...
    @property
    def generation(self)->int: ...
    @property
    def prompt_feeding(self)->int: ...
    
class GenerationResult():
    """
    Containes the result of a generation and the time taken for each step.
    """
    @property
    def text(self)->str: ...
    @property
    def times(self)->GenerationTimes: ...
    @property
    def stop_reason(self)->StopReason: ...


from .config import QuantizationType,ContainerType,SessionConfig
import os
import pathlib
from .models import Mpt,GptNeoX,GptJ,Gpt2,Bloom,Llama
from .base_model import Model
import logging
from typing import Optional, List, Union,Type,Dict
import os
from enum import Enum, auto
from dataclasses import dataclass
import json
from blake3 import blake3
from huggingface_hub import snapshot_download
from huggingface_hub.utils import validate_repo_id, HFValidationError
from huggingface_hub import cached_assets_path

class QuantizationVersions(Enum):
    Not_Quantized=auto()
    V1=auto()
    V2=auto()

class KnownModels(Enum):
    GptNeoX = auto()
    Mpt = auto()
    GptJ = auto()
    Gpt2 = auto()
    Bloom = auto()
    Llama = auto()


_QUANTIZATION_TYPE_MAP = {
    "Q4_0": QuantizationType.Q4_0,
    "Q4_1": QuantizationType.Q4_1,
    "Q5_0": QuantizationType.Q5_0,
    "Q5_1": QuantizationType.Q5_1,
    "Q8_0": QuantizationType.Q8_0,
    "F16": QuantizationType.F16
}

_CONTAINER_TYPE_MAP = {
    "GGML": ContainerType.GGML,
    "GGJT": ContainerType.GGJT
}

_KNOWN_MODELS_MAP = {
    KnownModels.GptNeoX: GptNeoX,
    KnownModels.Mpt: Mpt,
    KnownModels.GptJ: GptJ,
    KnownModels.Gpt2: Gpt2,
    KnownModels.Bloom: Bloom,
    KnownModels.Llama: Llama
}

CURRENT_QUANTIZATION_VERSION = QuantizationVersions.V2

class PathType(Enum):
    DIR = auto()
    FILE = auto()
    REPO = auto()
    UNKNOWN = auto() 

def _get_path_type(path: Union[str,os.PathLike]) -> PathType:
    p = pathlib.Path(path)
    if p.is_file():
        return PathType.FILE
    elif p.is_dir():
        return PathType.DIR
    try:
        validate_repo_id(str(path))
        return PathType.REPO
    except HFValidationError:
        pass
    return PathType.UNKNOWN


@dataclass()
class ModelMetadata():
    """
    A dataclass to store metadata about a model.
    """
    model: KnownModels
    quantization: QuantizationType
    container: ContainerType
    quantization_version: QuantizationVersions=QuantizationVersions.Not_Quantized
    converter:str="llm-rs"
    hash:Optional[str]=None
    base_model:Optional[str]=None

    def add_hash(self,model_path:Union[str,os.PathLike]):
        h  = blake3(max_threads=blake3.AUTO)
        b  = bytearray(128_000_000)
        mv = memoryview(b)
        with open(model_path, 'rb', buffering=0) as f:
            for n in iter(lambda : f.readinto(mv), 0):
                h.update(mv[:n])
        self.hash=h.hexdigest()

    def serialize(self):
        return {
            "model": self.model.name,
            "quantization": repr(self.quantization).split(".")[-1],
            "quantization_version": self.quantization_version.name,
            "container": repr(self.container).split(".")[-1],
            "converter": self.converter,
            "hash": self.hash,
            "base_model": self.base_model
        }
    
    @staticmethod
    def deserialize(metadata_dict:Dict[str,str])->"ModelMetadata":
        return ModelMetadata(
            model = KnownModels[metadata_dict["model"]],
            quantization = _QUANTIZATION_TYPE_MAP[metadata_dict["quantization"]],
            quantization_version= QuantizationVersions[metadata_dict["quantization_version"]],
            container = _CONTAINER_TYPE_MAP[metadata_dict["container"]],
            converter = metadata_dict["converter"],
            hash = metadata_dict.get("hash"),
            base_model = metadata_dict.get("base_model")
        )


@dataclass
class AutoConfig():
    repo_type:Optional[str] = None
    model_type: Optional[str] = None

    @classmethod
    def from_pretrained(
        cls,
        model_path_or_repo_id: Union[str, os.PathLike],
        **kwargs,
    ) -> 'AutoConfig':
        path_type = _get_path_type(model_path_or_repo_id)
        if path_type == PathType.UNKNOWN:
            raise ValueError(
                f"Model path '{model_path_or_repo_id}' doesn't exist.")
        elif path_type == PathType.FILE:
            raise ValueError(
                f"Model path '{model_path_or_repo_id}' is a file. "
                "Please provide a directory or a repo id.")
        
        auto_config = AutoConfig()
        if path_type == PathType.DIR:
            cls._update_from_dir(str(model_path_or_repo_id), auto_config)
        elif path_type == PathType.REPO:
            cls._update_from_repo(str(model_path_or_repo_id), auto_config)

        return auto_config

    @classmethod
    def _update_from_repo(
        cls,
        repo_id: str,
        auto_config: 'AutoConfig',
    ) -> None:
        path = snapshot_download(repo_id=repo_id, allow_patterns='config.json')
        cls._update_from_dir(path, auto_config)

    @classmethod
    def _update_from_dir(cls, path: str, auto_config: 'AutoConfig') -> None:
        resolved_path = (pathlib.Path(path) / 'config.json').resolve()
        if resolved_path.is_file():
            cls._update_from_file(str(resolved_path), auto_config)
        else:
            raise ValueError(f"Config path '{resolved_path}' doesn't exist.")

    @classmethod
    def _update_from_file(cls, path: str, auto_config: 'AutoConfig') -> None:
        with open(path) as f:
            config = json.load(f)
        auto_config.model_type = config.get('model_type')
        auto_config.repo_type = config.get('repo_type')

class AutoModel():
    """
    Utility to load models, without having to specify the model type.
    """
    @classmethod
    def load_metadata(cls,model_file:Union[str,os.PathLike])->ModelMetadata:
        path = pathlib.Path(model_file)
        if not path.is_file():
            raise ValueError(f"Model file '{model_file}' is not a file!")
        if not path.exists():
            raise ValueError(f"Model file '{model_file}' does not exist!")

        metadata_file = path.with_suffix(".meta")
        if not metadata_file.exists():
            raise ValueError(f"Model file '{model_file}' does not have a metadata file '{metadata_file}'! If you want to autoload this model, please specify the model type.")
        
        metadata = ModelMetadata.deserialize(json.loads(metadata_file.read_text()))
        return metadata
    
    @classmethod
    def _infer_model_type(cls,model_file:Union[str,os.PathLike],known_model:Optional[KnownModels]=None)->Type[Model]:
        model_to_lookup = None
        if known_model:
            model_to_lookup = known_model
        else:
            metadata = cls.load_metadata(model_file)
            model_to_lookup = metadata.model

        if model_to_lookup in _KNOWN_MODELS_MAP:
            return _KNOWN_MODELS_MAP[model_to_lookup]
        else:
            raise ValueError(f"Unknown model type '{model_to_lookup}'")
            
        
    @classmethod
    def from_file(cls, path:Union[str,os.PathLike],
                  model_type: Optional[KnownModels] = None,
                  session_config:SessionConfig=SessionConfig(),
                  lora_paths:Optional[List[Union[str,os.PathLike]]]=None,
                  verbose:bool=False)->Model:
        
        model = cls._infer_model_type(path,model_type)
        return model(path,session_config,lora_paths,verbose)
    
    @classmethod
    def from_pretrained(cls,
        model_path_or_repo_id: Union[str,os.PathLike],
        model_file: Optional[str] = None,
        model_type: Optional[KnownModels] = None,
        session_config:SessionConfig=SessionConfig(),
        lora_paths:Optional[List[Union[str,os.PathLike]]]=None,
        verbose:bool=False,
        default_quantization:QuantizationType=QuantizationType.Q4_0,
        default_container:ContainerType=ContainerType.GGJT)->Model:

        path_type = _get_path_type(model_path_or_repo_id)

        if path_type == PathType.UNKNOWN:
            raise ValueError(f"Unknown path type for '{model_path_or_repo_id}'")
        elif path_type == PathType.FILE:
            return cls.from_file(model_path_or_repo_id,model_type,session_config,lora_paths,verbose)
        else:
            if path_type == PathType.REPO:

                try: 
                    config = AutoConfig.from_pretrained(
                        model_path_or_repo_id,
                    )
                except ValueError:
                    logging.warning("Could not find config.json in repo, assuming GGML model...")
                    config = AutoConfig(repo_type="GGML")
                       

                if config.repo_type != "GGML":
                    logging.warning("Found normal HuggingFace model, starting conversion...")
                    return cls.from_transformer(model_path_or_repo_id, session_config, lora_paths, verbose, default_quantization, default_container)
            
                resolved_path = cls._find_model_path_from_repo(str(model_path_or_repo_id),model_file)
                return cls.from_file(resolved_path,model_type,session_config,lora_paths,verbose)
            
            elif path_type == PathType.DIR:
                resolved_path = cls._find_model_path_from_dir(str(model_path_or_repo_id),model_file)
                return cls.from_file(resolved_path,model_type,session_config,lora_paths,verbose)
            
            else:
                raise ValueError(f"Unknown path type '{path_type}'")


    @classmethod
    def _find_model_path_from_dir(
        cls,
        directory: str,
        filename: Optional[str] = None,
    ) -> str:
        path = pathlib.Path(directory).resolve()
        if filename:
            file = (path / filename)
            if not file.is_file():
                raise ValueError(
                    f"Model file '{filename}' not found in '{path}'")
            return str(file)

        files = [
            f for f in path.iterdir()
            if f.is_file() and f.name.endswith('.bin')
        ]

        if len(files) == 0:
            raise ValueError(f"No model files found in '{path}'")
        elif len(files) > 1:
            raise ValueError(
                f"Multiple model files found in '{path}'! Please specify one of the following model files to load: '{','.join([f.name for f in files])}'"
            )

        return str(files[0].resolve())
    
    @classmethod
    def _find_model_path_from_repo(
        cls,
        repo_id: str,
        filename: Optional[str] = None,
    ) -> str:
        cache_directory = cached_assets_path("llm-rs",namespace=repo_id)
        #we dont want to download the whole repo, just the metadata
        directory = snapshot_download(repo_id=repo_id,
                                allow_patterns='*.meta',
                                local_dir=cache_directory)
        path = pathlib.Path(directory).resolve()
        files = [
            f for f in path.iterdir()
            if f.is_file() and f.name.endswith('.meta')
        ]

        if len(files) == 0 and filename is None:
            raise ValueError(f"No model files found in '{repo_id}'! Either specify the model file via the `model_file` parameter or make sure the repository contains a *.meta files for each model.")
        if len(files) > 1 and filename is None:
            raise ValueError(
                f"Multiple model files found in '{repo_id}'! Please specify one of the following model files to load: '{','.join([f.name.replace('.meta','.bin') for f in files])}' "
            )
        #if we have a filename, we can just download that file
        filename = filename if filename else files[0].name.replace('.meta','.bin')
        model_directory = snapshot_download(repo_id=repo_id,
                                allow_patterns=filename,
                                local_dir=cache_directory)
        
        return cls._find_model_path_from_dir(model_directory, filename=filename)
    
    @classmethod
    def from_transformer(cls,
        model_path_or_repo_id: Union[str,os.PathLike],
        session_config:SessionConfig=SessionConfig(),
        lora_paths:Optional[List[Union[str,os.PathLike]]]=None,
        verbose:bool=False,
        default_quantization:QuantizationType=QuantizationType.Q4_0,
        default_container:ContainerType=ContainerType.GGJT):
        
        try:
            from .convert import AutoConverter
            from .convert.auto_converter import get_name_from_config
            from transformers import AutoConfig
        except ImportError:
            raise ImportError("Model conversion needs additional dependencies. Please install 'llm-rs[convert]' to use this functionality!")
        

        config = AutoConfig.from_pretrained(model_path_or_repo_id,trust_remote_code=True)
        model_name = get_name_from_config(config)
        export_path = cached_assets_path("llm-rs",namespace=model_name)
        converted_model = AutoConverter.convert(model_path_or_repo_id,export_path)
        if default_quantization != QuantizationType.F16:
            converted_model = AutoQuantizer.quantize(converted_model,quantization=default_quantization,container=default_container)
        return cls.from_file(converted_model,None,session_config,lora_paths,verbose)
    
# Hack to make the quantization type enum hashable
_APPENDIX_MAP = {
    QuantizationType.Q4_0.__repr__(): "q4_0",
    QuantizationType.Q4_1.__repr__(): "q4_1",
    QuantizationType.Q5_0.__repr__(): "q5_0",
    QuantizationType.Q5_1.__repr__(): "q5_1",
    QuantizationType.Q8_0.__repr__(): "q8_0",
}

class AutoQuantizer():
    """
    Utility to quantize models, without having to specify the model type.
    """
    @staticmethod
    def quantize(model_file:Union[str,os.PathLike],target_path:Optional[Union[str,os.PathLike]]=None,quantization:QuantizationType=QuantizationType.Q4_0,container:ContainerType=ContainerType.GGJT)->Union[str,os.PathLike]:
        metadata=AutoModel.load_metadata(model_file)
        if metadata.quantization != QuantizationType.F16:
            raise ValueError(f"Model '{model_file}' is already quantized to '{metadata.quantization}'")

        model_type = AutoModel._infer_model_type(model_file)

        if target_path is None:
            target_path = os.path.dirname(model_file)
            
        def build_target_name()->str:
            output_path = pathlib.Path(target_path)
            if output_path.is_file():
                return str(output_path)
            else:
                output_path.mkdir(parents=True,exist_ok=True)
                model_path = pathlib.Path(model_file)
                appendix = ""

                if quantization.__repr__() in _APPENDIX_MAP:
                    appendix += f"-{_APPENDIX_MAP[quantization.__repr__()]}"

                if container == ContainerType.GGJT:
                    appendix += "-ggjt"
            
                filename = model_path.stem.replace("-f16","") + appendix + model_path.suffix
                return str(output_path / filename)

        target_file = build_target_name()
        if pathlib.Path(target_file).exists():
            logging.warning(f"Target file '{target_file}' already exists, skipping quantization")
            return target_file
        
        logging.info(f"Quantizing model '{model_file}' to '{target_file}'")
        model_type.quantize(str(model_file),target_file,quantization,container)

        metadata_file = pathlib.Path(target_file).with_suffix(".meta")
        quantized_metadata = ModelMetadata(model=metadata.model,quantization=quantization,container=container,quantization_version=CURRENT_QUANTIZATION_VERSION,base_model=metadata.base_model)
        quantized_metadata.add_hash(target_file)
        logging.info(f"Writing metadata file '{metadata_file}'")
        metadata_file.write_text(json.dumps(quantized_metadata.serialize(),indent=4))
        logging.info(f"Finished quantizing model '{model_file}' to '{target_file}'")
        return target_file





from ._base import BaseAdapter
from typing import Union,Tuple,Optional,BinaryIO 
from transformers import AutoTokenizer,AutoModelForCausalLM,AutoConfig
import os
import torch
import struct
from ...auto import KnownModels

#based on https://github.com/ggerganov/ggml/blob/master/examples/mpt/convert-h5-to-ggml.py
class MPTConverter(BaseAdapter):
    model_type:KnownModels=KnownModels.Mpt
    
    def load(self,pretrained_model_name_or_path:Union[str,os.PathLike],pretrained_tokenizer_name_or_path:Optional[Union[str,os.PathLike]]=None)->Tuple[AutoTokenizer,AutoModelForCausalLM]:
        config = AutoConfig.from_pretrained(pretrained_model_name_or_path,trust_remote_code=True)
        
        model = AutoModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path,
            torch_dtype=torch.float16,
            trust_remote_code=True,
            low_cpu_mem_usage=True,
            )
        
        #MPT uses the `EleutherAI/gpt-neox-20b` tokenizer => use it if no tokenizer is specified
        tokenizer_name = pretrained_tokenizer_name_or_path if pretrained_tokenizer_name_or_path else "EleutherAI/gpt-neox-20b" 
        tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)
        return config,tokenizer,model
    
    def _write_hyperparameters(self,out_file:BinaryIO):
        hyperparameters = self.config.to_dict()
        out_file.write(struct.pack("i", hyperparameters["d_model"]))
        out_file.write(struct.pack("i", hyperparameters["max_seq_len"]))
        out_file.write(struct.pack("i", hyperparameters["n_heads"]))
        out_file.write(struct.pack("i", hyperparameters["n_layers"]))
        out_file.write(struct.pack("i", hyperparameters["vocab_size"]))
        out_file.write(struct.pack("f", hyperparameters["attn_config"]["alibi_bias_max"]))
        out_file.write(struct.pack("f", hyperparameters["attn_config"]["clip_qkv"] or 0.0))


import numpy as np
from ._base import BaseAdapter
from typing import Union,Tuple,Optional,BinaryIO 
from transformers import AutoTokenizer,AutoModelForCausalLM,AutoConfig,GPTJForCausalLM
import os
import torch
import struct
import numpy as np
from ...auto import KnownModels

#based on https://github.com/ggerganov/ggml/blob/master/examples/gpt-j/convert-h5-to-ggml.py
class GptJConverter(BaseAdapter):
    model_type:KnownModels=KnownModels.GptJ

    def load(self,pretrained_model_name_or_path:Union[str,os.PathLike],pretrained_tokenizer_name_or_path:Optional[Union[str,os.PathLike]]=None)->Tuple[AutoTokenizer,AutoModelForCausalLM]:
        config = AutoConfig.from_pretrained(pretrained_model_name_or_path)
        
        model = GPTJForCausalLM.from_pretrained(
            pretrained_model_name_or_path,
            torch_dtype=torch.float16,
            low_cpu_mem_usage=True,
            )
        tokenizer_name = pretrained_tokenizer_name_or_path if pretrained_tokenizer_name_or_path else pretrained_model_name_or_path 
        tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)
        return config,tokenizer,model
    
    def _write_hyperparameters(self,out_file:BinaryIO):
        hyperparameters = self.config.to_dict()
        out_file.write(struct.pack("i", hyperparameters["vocab_size"]))
        out_file.write(struct.pack("i", hyperparameters["n_positions"]))
        out_file.write(struct.pack("i", hyperparameters["n_embd"]))
        out_file.write(struct.pack("i", hyperparameters["n_head"]))
        out_file.write(struct.pack("i", hyperparameters["n_layer"]))
        out_file.write(struct.pack("i", hyperparameters["rotary_dim"]))

    def _write_vocabulary(self, out_file: BinaryIO):
        # write the vocabulary size
        out_file.write(struct.pack("i", self.config.vocab_size))
        return super()._write_vocabulary(out_file)
    

    def _filter_weights(self, name: str, weight: torch.Tensor) -> bool:
        return name.endswith("attn.masked_bias") or name.endswith(".attn.bias") 


from ._base import BaseAdapter,GGJT_MAGIC
from typing import Union,Tuple,Optional,BinaryIO,List,Iterable
from transformers import LlamaConfig,LlamaForCausalLM,LlamaTokenizer,AutoConfig,AutoTokenizer,AutoModelForCausalLM
import os
import torch
import struct
from ...auto import KnownModels
import re
import numpy as np
import logging

#based on https://github.com/ggerganov/llama.cpp/blob/master/convert.py
class LlamaConverter(BaseAdapter):
    model_type:KnownModels=KnownModels.Llama
    file_magic=GGJT_MAGIC

    def load(self,pretrained_model_name_or_path:Union[str,os.PathLike],pretrained_tokenizer_name_or_path:Optional[Union[str,os.PathLike]]=None)->Tuple[AutoConfig,AutoTokenizer,AutoModelForCausalLM]:
        config = LlamaConfig.from_pretrained(pretrained_model_name_or_path)
        
        model = LlamaForCausalLM.from_pretrained(
            pretrained_model_name_or_path,
            torch_dtype=torch.float16,
            low_cpu_mem_usage=True,
            )
        
        tokenizer_name = pretrained_tokenizer_name_or_path if pretrained_tokenizer_name_or_path else pretrained_model_name_or_path 
        tokenizer = LlamaTokenizer.from_pretrained(tokenizer_name)

        def make_tensors_list() -> List[str]:
            ret = [
                'tok_embeddings.weight',
                'norm.weight',
                'output.weight',
            ]
            for i in range(80):  # maximum number of layer
                ret += [
                    f'layers.{i}.attention.wq.weight',
                    f'layers.{i}.attention.wk.weight',
                    f'layers.{i}.attention.wv.weight',
                    f'layers.{i}.attention.wo.weight',
                    f'layers.{i}.attention_norm.weight',
                    f'layers.{i}.feed_forward.w1.weight',
                    f'layers.{i}.feed_forward.w2.weight',
                    f'layers.{i}.feed_forward.w3.weight',
                    f'layers.{i}.ffn_norm.weight',
                ]
            return ret

        self.TENSORS_LIST = make_tensors_list()
        self.TENSORS_SET = set(self.TENSORS_LIST)

        self.mapping = {
            "model.layers.{i}.self_attn.q_proj.weight": "layers.{i}.attention.wq.weight",
            "model.layers.{i}.self_attn.k_proj.weight": "layers.{i}.attention.wk.weight",
            "model.layers.{i}.self_attn.v_proj.weight": "layers.{i}.attention.wv.weight",
            "model.layers.{i}.self_attn.o_proj.weight": "layers.{i}.attention.wo.weight",
            "model.layers.{i}.mlp.gate_proj.weight": "layers.{i}.feed_forward.w1.weight",
            "model.layers.{i}.mlp.down_proj.weight": "layers.{i}.feed_forward.w2.weight",
            "model.layers.{i}.mlp.up_proj.weight": "layers.{i}.feed_forward.w3.weight",
            "model.layers.{i}.input_layernorm.weight": "layers.{i}.attention_norm.weight",
            "model.layers.{i}.post_attention_layernorm.weight": "layers.{i}.ffn_norm.weight",
        }

        return config,tokenizer,model
    
    def _write_hyperparameters(self,out_file:BinaryIO):
        hyperparameters = self.config.to_dict()
        out_file.write(struct.pack("i", hyperparameters["vocab_size"]))
        out_file.write(struct.pack("i", hyperparameters["hidden_size"]))
        out_file.write(struct.pack("i", 256))
        out_file.write(struct.pack("i", hyperparameters["num_attention_heads"]))
        out_file.write(struct.pack("i", hyperparameters["num_hidden_layers"]))
        out_file.write(struct.pack("i", hyperparameters["hidden_size"]//hyperparameters["num_attention_heads"])) # rot (obsolete)

    def _rename_weights(self, name: str) -> str:
        if name == "model.embed_tokens.weight":
            return "tok_embeddings.weight"
        elif name == "model.norm.weight":
            return "norm.weight"
        elif name == "lm_head.weight":
            return "output.weight"
        
        for old, new in self.mapping.items():
            pattern = old.replace("{i}", "(\d+)")
            match = re.fullmatch(pattern, name)
            if match:
                return new.replace("{i}", match.group(1))
            
        return name
    
    def _transform_weights(self, name: str, weight: torch.Tensor) -> torch.Tensor:
        to_permut = [r"layers.(\d+).attention.wq.weight",r"layers.(\d+).attention.wk.weight"]
        for pattern in to_permut:
            match = re.fullmatch(pattern, name)
            if match:
                n_heads = self.config.num_attention_heads
                numpy_weights:np.ndarray = weight.numpy()
                reshaped = (numpy_weights.reshape(n_heads, 2, numpy_weights.shape[0] // n_heads // 2, *numpy_weights.shape[1:])
                   .swapaxes(1, 2)
                   .reshape(numpy_weights.shape))
                logging.info(f"Permuting {name} from {weight.shape} to {reshaped.shape}")
                return torch.from_numpy(reshaped)
        return weight
    
    def _filter_weights_after_rename(self, name: str, weight: torch.Tensor) -> bool:
            return name not in self.TENSORS_SET
    

    def _write_vocabulary(self,out_file:BinaryIO):
        sentence_piece_tokenizer = self.tokenizer.sp_model

        logging.info(f"Processing vocabulary with size {self.config.vocab_size}")
        def sentencepiece_tokens() -> Iterable[Tuple[bytes, float]]:
            tokenizer = sentence_piece_tokenizer
            for i in range(tokenizer.vocab_size()):
                text: bytes
                if tokenizer.is_unknown(i):
                    text = " \u2047 ".encode("utf-8")
                elif tokenizer.is_control(i):
                    text = b""
                elif tokenizer.is_byte(i):
                    piece = tokenizer.id_to_piece(i)
                    if len(piece) != 6:
                        raise Exception(f"Invalid token: {piece}")
                    byte_value = int(piece[3:-1], 16)
                    text = struct.pack("B", byte_value)
                else:
                    text = tokenizer.id_to_piece(i).replace("\u2581", " ").encode("utf-8")
                score: float = tokenizer.get_score(i)
                yield text, score

        for text, score in sentencepiece_tokens():
            out_file.write(struct.pack("i", len(text)))
            out_file.write(text)
            out_file.write(struct.pack("f", score))

    

